@@ -1,6 +1,6 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /*
- * Copyright (c) 2008  LiTL, LLC
+ * Copyright (c) 2008  litl, LLC
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -28,8 +28,7 @@
 #include "boxed.h"
 #include "arg.h"
 #include "object.h"
-#include <gjs/mem.h>
-#include <gjs/jsapi-util.h>
+#include <gjs/gjs.h>
 #include "repo.h"
 #include "function.h"
 
@@ -101,6 +100,16 @@ boxed_new_resolve(JSContext *context,
 
             method_name = g_base_info_get_name( (GIBaseInfo*) method_info);
 
+            /* we do not define deprecated methods in the prototype */
+            if (g_base_info_is_deprecated( (GIBaseInfo*) method_info)) {
+                gjs_debug(GJS_DEBUG_GBOXED,
+                          "Ignoring definition of deprecated method %s in prototype %s.%s",
+                          method_name,
+                          g_base_info_get_namespace( (GIBaseInfo*) priv->info),
+                          g_base_info_get_name( (GIBaseInfo*) priv->info));
+                return JS_FALSE;
+            }
+
             gjs_debug(GJS_DEBUG_GBOXED,
                       "Defining method %s in prototype for %s.%s",
                       method_name,
@@ -202,6 +211,9 @@ boxed_constructor(JSContext *context,
     JSObject *proto;
     gboolean is_proto;
 
+    if (!gjs_check_constructing(context))
+        return JS_FALSE;
+
     priv = g_slice_new0(Boxed);
 
     GJS_INC_COUNTER(boxed);
@@ -534,7 +546,8 @@ gjs_boxed_from_g_boxed(JSContext    *context,
         return NULL;
     }
 
-    if (g_base_info_get_type( (GIBaseInfo*) info) != GI_INFO_TYPE_BOXED) {
+    if (g_base_info_get_type( (GIBaseInfo*) info) != GI_INFO_TYPE_BOXED &&
+        g_base_info_get_type( (GIBaseInfo*) info) != GI_INFO_TYPE_STRUCT) {
         gjs_throw(context,
                   "GType %s doesn't map to boxed in g-i?",
                   g_base_info_get_name( (GIBaseInfo*) info));
