@@ -1,6 +1,6 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /*
- * Copyright (c) 2008  LiTL, LLC
+ * Copyright (c) 2008  litl, LLC
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -24,7 +24,6 @@
 #include <config.h>
 
 #include <util/log.h>
-#include <util/dirs.h>
 #include <util/glib.h>
 
 #include <jsapi.h>
@@ -36,6 +35,8 @@
 
 #include <string.h>
 
+static char **gjs_search_path = NULL;
+
 typedef struct {
     void *dummy;
 } Importer;
@@ -130,18 +131,15 @@ finish_import(JSContext  *context,
 }
 
 static JSBool
-finish_import_and_define(JSContext  *context,
-                         JSObject   *obj,
-                         JSObject   *module_obj,
-                         const char *name)
+define_import(JSContext  *context,
+              JSObject   *obj,
+              JSObject   *module_obj,
+              const char *name)
 {
-    if (!finish_import(context, name))
-        return JS_FALSE;
-
     if (!JS_DefineProperty(context, obj,
                            name, OBJECT_TO_JSVAL(module_obj),
                            NULL, NULL,
-                           GJS_MODULE_PROP_FLAGS)) {
+                           GJS_MODULE_PROP_FLAGS & ~JSPROP_PERMANENT)) {
         gjs_debug(GJS_DEBUG_IMPORTER,
                   "Failed to define '%s' in importer",
                   name);
@@ -151,6 +149,59 @@ finish_import_and_define(JSContext  *context,
     return JS_TRUE;
 }
 
+/* Just like define_import, but makes the setting permament.
+ * we do this after the import succesfully completes.
+ */
+static void
+seal_import(JSContext    *context,
+            JSObject   *obj,
+            JSObject   *module_obj,
+            const char *name)
+{
+    if (!JS_DefineProperty(context, obj,
+                           name, OBJECT_TO_JSVAL(module_obj),
+                           NULL, NULL,
+                           GJS_MODULE_PROP_FLAGS)) {
+        gjs_debug(GJS_DEBUG_IMPORTER,
+                  "Failed to permanently define '%s' in importer",
+                  name);
+    }
+}
+
+/* An import failed. Delete the property pointing to the import
+ * from the parent namespace. In complicated situations this might
+ * not be sufficient to get us fully back to a sane state. If:
+ *
+ *  - We import module A
+ *  - module A imports module B
+ *  - module B imports module A, storing a reference to the current
+ *    module A module object
+ *  - module A subsequently throws an exception
+ *
+ * Then module B is left imported, but the imported module B has
+ * a reference to the failed module A module object. To handle this
+ * we could could try to track the entire "import operation" and
+ * roll back *all* modifications made to the namespace objects.
+ * It's not clear that the complexity would be worth the small gain
+ * in robustness. (You can still come up with ways of defeating
+ * the attempt to clean up.)
+ */
+static void
+cancel_import(JSContext  *context,
+              JSObject   *obj,
+              const char *name)
+{
+    gjs_debug(GJS_DEBUG_IMPORTER,
+              "Cleaning up from failed import of '%s'",
+              name);
+
+    if (!JS_DeleteProperty(context, obj, name)) {
+        gjs_debug(GJS_DEBUG_IMPORTER,
+                  "Failed to delete '%s' in importer",
+                  name);
+    }
+}
+
 static JSBool
 import_native_file(JSContext  *context,
                    JSObject   *obj,
@@ -169,7 +220,13 @@ import_native_file(JSContext  *context,
         return JS_FALSE;
     }
 
-    JS_AddRoot(context, &module_obj);
+    /* We store the module object into the parent module before
+     * initializing the module. If the module has the
+     * GJS_NATIVE_SUPPLIES_MODULE_OBJ flag, it will just overwrite
+     * the reference we stored when it initializes.
+     */
+    if (!define_import(context, obj, module_obj, name))
+        return JS_FALSE;
 
     if (!define_meta_properties(context, module_obj, name, obj))
         goto out;
@@ -177,21 +234,16 @@ import_native_file(JSContext  *context,
     if (!gjs_import_native_module(context, module_obj, full_path, &flags))
         goto out;
 
-    if (flags & GJS_NATIVE_SUPPLIES_MODULE_OBJ) {
-        /* In this case module_obj just gets garbage collected,
-         * we don't end up using it.
-         */
-        if (!finish_import(context, name))
-            goto out;
-    } else {
-        if (!finish_import_and_define(context, obj, module_obj, name))
-            goto out;
-    }
+    if (!finish_import(context, name))
+        goto out;
 
     retval = JS_TRUE;
 
  out:
-    JS_RemoveRoot(context, &module_obj);
+    if (!retval)
+        cancel_import(context, obj, name);
+    else if (!(flags & GJS_NATIVE_SUPPLIES_MODULE_OBJ))
+        seal_import(context, obj, module_obj, name);
 
     return retval;
 }
@@ -206,7 +258,8 @@ import_file(JSContext  *context,
     gsize script_len;
     JSObject *module_obj;
     GError *error;
-    jsval retval;
+    jsval script_retval;
+    JSBool retval = JS_FALSE;
 
     gjs_debug(GJS_DEBUG_IMPORTER,
               "Importing '%s'", full_path);
@@ -216,9 +269,12 @@ import_file(JSContext  *context,
         return JS_FALSE;
     }
 
-    if (!define_meta_properties(context, module_obj, name, obj))
+    if (!define_import(context, obj, module_obj, name))
         return JS_FALSE;
 
+    if (!define_meta_properties(context, module_obj, name, obj))
+        goto out;
+
     script = NULL;
     script_len = 0;
 
@@ -226,7 +282,7 @@ import_file(JSContext  *context,
     if (!g_file_get_contents(full_path, &script, &script_len, &error)) {
         gjs_throw(context, "Could not open %s: %s", full_path, error->message);
         g_error_free(error);
-        return JS_FALSE;
+        goto out;
     }
 
     g_assert(script != NULL);
@@ -237,7 +293,7 @@ import_file(JSContext  *context,
                            script_len,
                            full_path,
                            1, /* line number */
-                           &retval)) {
+                           &script_retval)) {
         g_free(script);
 
         /* If JSOPTION_DONT_REPORT_UNCAUGHT is set then the exception
@@ -254,15 +310,23 @@ import_file(JSContext  *context,
                          "JS_EvaluateScript() returned FALSE but did not set exception");
         }
 
-        return JS_FALSE;
+        goto out;
     }
 
     g_free(script);
 
-    if (!finish_import_and_define(context, obj, module_obj, name))
-        return JS_FALSE;
+    if (!finish_import(context, obj))
+        goto out;
 
-    return JS_TRUE;
+    retval = JS_TRUE;
+
+ out:
+    if (!retval)
+        cancel_import(context, obj, name);
+    else
+        seal_import(context, obj, module_obj, name);
+
+    return retval;
 }
 
 static JSBool
@@ -329,6 +393,9 @@ do_import(JSContext  *context,
             goto out;
         }
 
+        g_free(dirname);
+        dirname = NULL;
+
         if (!gjs_string_to_utf8(context, elem, &dirname))
             goto out; /* Error message already set */
 
@@ -565,7 +632,13 @@ importer_new_enumerate(JSContext  *context,
             }
 
             while ((filename = g_dir_read_name(dir))) {
-                char *full_path = g_build_filename(dirname, filename, NULL);
+                char *full_path;
+
+                /* skip hidden files and directories (.svn, .git, ...) */
+                if (filename[0] == '.')
+                    continue;
+
+                full_path = g_build_filename(dirname, filename, NULL);
 
                 if (g_file_test(full_path, G_FILE_TEST_IS_DIR)) {
                     g_ptr_array_add(iter->elements, g_strdup(filename));
@@ -820,6 +893,57 @@ importer_new(JSContext    *context)
     return importer;
 }
 
+static G_CONST_RETURN char * G_CONST_RETURN *
+gjs_get_search_path(void)
+{
+    char **search_path;
+
+    /* not thread safe */
+
+    if (!gjs_search_path) {
+        G_CONST_RETURN gchar* G_CONST_RETURN * system_data_dirs;
+        const char *envstr;
+        GPtrArray *path;
+        gsize i;
+
+        path = g_ptr_array_new();
+
+        /* in order of priority */
+
+        /* $GJS_PATH */
+        envstr = g_getenv("GJS_PATH");
+        if (envstr) {
+            char **dirs, **d;
+            dirs = g_strsplit(envstr, G_SEARCHPATH_SEPARATOR_S, 0);
+            for (d = dirs; *d != NULL; d++)
+                g_ptr_array_add(path, *d);
+            /* we assume the array and strings are allocated separately */
+            g_free(dirs);
+        }
+
+        /* $XDG_DATA_DIRS /gjs-1.0 */
+        system_data_dirs = g_get_system_data_dirs();
+        for (i = 0; system_data_dirs[i] != NULL; ++i) {
+            char *s;
+            s = g_build_filename(system_data_dirs[i], "gjs-1.0", NULL);
+            g_ptr_array_add(path, s);
+        }
+
+        /* ${libdir}/gjs-1.0 */
+        g_ptr_array_add(path, g_strdup(GJS_NATIVE_DIR));
+
+        g_ptr_array_add(path, NULL);
+
+        search_path = (char**)g_ptr_array_free(path, FALSE);
+
+        gjs_search_path = search_path;
+    } else {
+        search_path = gjs_search_path;
+    }
+
+    return (G_CONST_RETURN char * G_CONST_RETURN *)search_path;
+}
+
 JSObject*
 gjs_define_importer(JSContext    *context,
                     JSObject     *in_object,
@@ -828,19 +952,16 @@ gjs_define_importer(JSContext    *context,
                     gboolean      add_standard_search_path)
 {
     JSObject *importer;
-    char **paths[3] = {0};
+    char **paths[2] = {0};
     char **search_path;
 
     paths[0] = (char**)initial_search_path;
     if (add_standard_search_path) {
         /* Stick the "standard" shared search path after the provided one. */
-        paths[1] = gjs_get_search_path(GJS_DIRECTORY_SHARED_JAVASCRIPT);
-        paths[2] = gjs_get_search_path(GJS_DIRECTORY_SHARED_JAVASCRIPT_NATIVE);
+        paths[1] = (char**)gjs_get_search_path();
     }
 
-    search_path = gjs_g_strv_concat(paths, 3);
-    g_strfreev(paths[1]);
-    g_strfreev(paths[2]);
+    search_path = gjs_g_strv_concat(paths, 2);
 
     importer = importer_new(context);
 
